"""
The `analytic_functions.py` module provides various analytic functions for processing
and analyzing comments data. It includes methods to calculate replies, votes, valence,
extremity, and Bayesian-corrected measures.

"""

from typing import Tuple
import numpy as np
import pandas as pd


class AnalyticFunctions:
    """
    A class used to encapsulate analytic functions for processing comment data.

    Methods
    -------
    calculate_number_of_replies(replies: list) -> int
        Calculate the number of replies.
    calculate_totalvotes(upvotes: int, downvotes: int) -> int
        Calculate the total number of votes.
    calculate_valence(downvotes: int, totalvotes: int) -> float
        Calculate the valence of the comments.
    calculate_extremity(valence: float) -> float
        Calculate the extremity of the comments.
    calculate_bayes_correction_general(measure: float, volume: float, weight: float, average_measure: float) -> float
        Calculate the Bayesian corrected value for a general measure.
    calculate_bayes_corrected_value_for_quantile_weights(measure: float, totalvotes: int, weights_for_quantiles: dict[str, float]) -> Tuple[float, float, float]
        Calculate Bayesian corrected values for different quantile weights.
    get_values_for_quantiles_of_measure(measure_data: pd.Series) -> dict[str, float]
        Get values for quantiles of a given measure.
    """

    def __int__(self):
        """
        Constructs the AnalyticFunctions object.
        Only passes as the class is just used to encapsulate analytic functions.
        """
        pass

    @staticmethod
    def calculate_number_of_replies(replies: list) -> int:
        """
        Calculate the number of replies from a reply list.

        Parameters
        ----------
        replies : list
            List containing replies.

        Returns
        -------
        int
            The number of replies.
        """
        number_of_replies: int = len(replies)
        return number_of_replies

    @staticmethod
    def calculate_totalvotes(upvotes: int, downvotes: int) -> int:
        """
        Calculate the total number of votes.

        Parameters
        ----------
        upvotes : int
            Number of upvotes.
        downvotes : int
            Number of downvotes.

        Returns
        -------
        int
            The total number of votes.
        """
        totalvotes: int = upvotes + downvotes
        return totalvotes

    @staticmethod
    def calculate_valence(downvotes: int, totalvotes: int) -> float:
        """
        Calculate the valence (overall attitude of votes) of a comment.

        Parameters
        ----------
        downvotes : int
            Number of downvotes on the comment.
        totalvotes : int
            Total number of votes on the comment.

        Returns
        -------
        float
            The valence of the comment.
        """
        if totalvotes == 0:
            return float("NaN")
        else:
            valence: float = -1 * (downvotes / totalvotes - 0.5)
        return valence

    @staticmethod
    def calculate_extremity(valence: float) -> float:
        """
        Calculate the extremity of a comment.
        The extremity is the absolute value of the valence.

        Parameters
        ----------
        valence : float
            The valence of the comment.

        Returns
        -------
        float
            The extremity of the comment.
        """
        if valence == float("NaN"):
            return float("NaN")
        else:
            extremity: float = abs(valence)
        return extremity

    @staticmethod
    def calculate_bayes_correction_general(
        measure: float, volume: float, weight: float, average_measure: float
    ) -> float:
        """
        Calculate the Bayesian corrected value for a general measure,
        correcting for volume of measurements a value is based on.

        Parameters
        ----------
        measure : float
            The original measure.
        volume : float
            The volume of data points.
        weight : float
            The weight to adjust the measure.
        average_measure : float
            The average measure used for correction.

        Returns
        -------
        float
            The Bayesian corrected value.
        """
        if volume == 0:
            return np.nan
        bayes_corrected_value: float = (
            volume / (volume + weight) * measure
            + weight / (volume + weight) * average_measure
        )
        return bayes_corrected_value

    def calculate_bayes_corrected_value_for_quantile_weights(
        self,
        measure: float,
        totalvotes: int,
        weights_for_quantiles: dict[str, float],
    ) -> Tuple[float, float, float]:
        """
        Calculate Bayesian corrected values for different quantile weights.

        Parameters
        ----------
        measure : float
            The original measure.
        totalvotes : int
            The total number of votes.
        weights_for_quantiles : dict[str, float]
            Dictionary containing weights for different quantiles.

        Returns
        -------
        Tuple[float, float, float]
            Bayesian corrected values for Q25, Q50, and Q75 quantiles.
        """
        weight_q25: float = weights_for_quantiles["weight_q25"]
        weight_q5: float = weights_for_quantiles["weight_q5"]
        weight_q75: float = weights_for_quantiles["weight_q75"]

        average_measure: float = weight_q5

        bayes_corrected_q25_measure: float = self.calculate_bayes_correction_general(
            measure, totalvotes, weight_q25, average_measure
        )
        bayes_corrected_q5_measure: float = self.calculate_bayes_correction_general(
            measure, totalvotes, weight_q5, average_measure
        )
        bayes_corrected_q75_measure: float = self.calculate_bayes_correction_general(
            measure, totalvotes, weight_q75, average_measure
        )

        return (
            bayes_corrected_q25_measure,
            bayes_corrected_q5_measure,
            bayes_corrected_q75_measure,
        )

    @staticmethod
    def get_values_for_quantiles_of_measure(
        measure_data: pd.Series,
    ) -> dict[str, float]:
        """
        Get values for quantiles of a given measure.

        Parameters
        ----------
        measure_data : pd.Series
            The measure data as a pandas Series.

        Returns
        -------
        dict[str, float]
            Dictionary containing values for Q25, Q50, and Q75 quantiles.
        """
        value_q25: float = measure_data.quantile(q=0.25)
        value_q5: float = measure_data.mean()
        value_q75: float = measure_data.quantile(q=0.75)

        results: dict = {
            "weight_q25": value_q25,
            "weight_q5": value_q5,
            "weight_q75": value_q75,
        }

        return results
