"""
The `comment_data_preprocessor.py` module provides a subclass of `DataPreprocessor`,
specifically for processing comment data files.

"""

from pathlib import Path
import pandas as pd

from src.data_preprocessor import DataPreprocessor
from src.single_comment_file_preprocessor import SingleCommentFilePreprocessor


class CommentsDataPreprocessor(DataPreprocessor):
    """
    A class used to preprocess comment data.

    Attributes
    ----------
    Inherits all attributes from the abstract base class DataPreprocessor.
    single_comment_file_preprocessor : SingleCommentFilePreprocessor
        An instance of the SingleCommentFilePreprocessor class which is used to preprocess individual comment files.

    Methods
    -------
    preprocess(comment_file_path: Path) -> pd.DataFrame
        Preprocess a comment file and return it as a DataFrame.
    """

    def __init__(self, data_folder_comment_data: Path):
        """
        Constructs the CommentsDataPreprocessor object.

        Parameters
        ----------
        data_folder_comment_data : Path
            Path object representing the directory where the comment data files are located.
        """
        super().__init__(data_folder_comment_data, split_word="SPON_comment")
        self.single_comment_file_preprocessor: SingleCommentFilePreprocessor = (
            SingleCommentFilePreprocessor()
        )

    def preprocess(self, comment_file_path: Path) -> pd.DataFrame:
        """
        Preprocess a comment data file and return it as a DataFrame.

        Parameters
        ----------
        comment_file_path : pathlib.Path
            The path object representing the comment data file to be preprocessed.

        Returns
        -------
        pd.DataFrame
            A pandas DataFrame containing the preprocessed comment data.
        """
        return self.single_comment_file_preprocessor.preprocess(comment_file_path)
