"""
The `create_analysis_dataset.py` module merges, processes and saves comment and article data into a final dataset.

This module provides the AnalysisDatasetCreator class, which performs tasks such as Bayesian correction, merging data,
removing irrelevant information, and saving the final analysis dataset.
"""

import numpy as np
import pandas as pd
from datetime import datetime
from pathlib import Path

from src.utils import save_data
from src.analytic_functions import AnalyticFunctions


class AnalysisDatasetCreator:
    """
    A class used to create an analysis dataset from comment data and article data.

    Attributes
    ----------
    comment_data : pd.DataFrame
        A dataframe containing comment data.
    article_data : pd.DataFrame
        A dataframe containing article data.
    add_extend_article_information : bool
        Boolean flag to decide whether to add extended information.
    name_dataset : str
        The name of the dataset to be created.
    analytic_functions : AnalyticFunctions
        An instance of the class AnalyticFunctions.
    calculate_bayes_corrected_value_for_quantile_weights_vectorized : np.vectorize
        A numpy vectorized function for calculating Bayesian corrected values.

    Methods
    -------
    process_comments_and_articles_into_analysis_dataset()
        Process comments and articles into a dataset ready for analysis.
    _merge_comments_and_article_data() -> pd.DataFrame
        Merge comment and article datasets.
    _process_comment_data()
        Process comment data.
    _process_article_data()
        Process article data.
    _calculate_bayesian_corrections()
        Calculate Bayesian corrections.
    _calculate_bayesian_corrected_measure(measure_name: str)
        Calculate Bayesian corrected measures.
    _get_bayesian_corrected_column_names(measure_name: str)
        Get the Bayesian-corrected column names.
    _remove_irrelevant_information_comments()
        Remove irrelevant information from comments.
    _set_comment_data_datatypes_for_efficient_storing()
        Set comment data datatypes for efficient storage.
    """

    def __init__(
        self,
        comment_data: pd.DataFrame,
        article_data: pd.DataFrame,
        add_extend_article_information: bool = True,
        add_comments_text_body: bool = True,
        name_dataset: Path = Path(
            f"{datetime.now().strftime('%Y-%m-%d')}_full_data.parquet"
        ),
    ):
        """
        Constructs the AnalysisDatasetCreator object.

        Parameters
        ----------
        comment_data : pd.DataFrame
            A dataframe containing the comment data.
        article_data : pd.DataFrame
            A dataframe containing the article data.
        add_extend_article_information : bool
            Flag to decide whether to add extended information.
        add_comments_text_body: bool
            Flag to decide whether to add the text of comments (memory intensive!)
        name_dataset : str
            The name of the dataset to be created.
        """
        self.comment_data: pd.DataFrame = comment_data
        self.article_data: pd.DataFrame = article_data
        self.add_extend_article_information: bool = add_extend_article_information
        self.add_comments_text_body: bool = add_comments_text_body
        self.name_dataset: Path = name_dataset
        self.analytic_functions: AnalyticFunctions = AnalyticFunctions()
        self.calculate_bayes_corrected_value_for_quantile_weights_vectorized = np.vectorize(
            self.analytic_functions.calculate_bayes_corrected_value_for_quantile_weights
        )

    def process_comments_and_articles_into_analysis_dataset(self):
        """
        Process comment and article data to create a dataset ready for analysis.

        This method sequentially calls other methods to process comment data, process article data,
        merge them, and finally save the merged dataset.
        """
        self._process_comment_data()
        self._process_article_data()
        merged_analysis_dataset: pd.DataFrame = self._merge_comments_and_article_data()
        save_data(self.name_dataset, merged_analysis_dataset)

    def _merge_comments_and_article_data(self) -> pd.DataFrame:
        """
        Merges the comment and article dataframes on the 'article' field.

        Returns
        -------
        pd.DataFrame
            A dataframe containing the merged comment and article data.
        """
        merged_dataset: pd.DataFrame = pd.merge(
            self.comment_data, self.article_data, on="article", how="left"
        )

        merged_dataset.rename(
            columns={
                "editing.edited": "edited",
                "channel": "section",
                "user.id": "user_id",
            },
            inplace=True,
        )

        merged_dataset: pd.DataFrame = self._calculate_derived_time_information(
            merged_dataset
        )

        return merged_dataset

    def _process_comment_data(self):
        """
        Perform pre-processing steps to comment data, includes removing irrelevant information
        and applying Bayesian corrections.
        """
        self._calculate_bayesian_corrections()
        self.comment_data.rename(
            columns={
                "mean_valence_replies": "mean valence of replies",
                "mean_extremity_replies": "mean extremity of replies",
            },
            inplace=True,
        )
        self._set_comment_data_datatypes_for_efficient_storing()
        self._remove_irrelevant_information_comments()

    def _process_article_data(self):
        """
        Perform pre-processing steps on article data, includes removing irrelevant information.
        """
        if self.add_extend_article_information:
            irrelevant_columns = [
                "url",
                "id",
                "comments_enabled",
                "date_modified",
                "date_published",
            ]
            self.article_data.drop(columns=irrelevant_columns, inplace=True)
        else:
            relevant_columns = ["article", "channel", "date_created"]
            self.article_data = self.article_data[relevant_columns]

    def _calculate_bayesian_corrections(self):
        """
        Apply bayesian corrections to valence, extremity and the mean valence and extremity of replies.
        """
        self._calculate_bayesian_corrected_measure("valence")
        self._calculate_bayesian_corrected_measure("extremity")
        self._calculate_bayesian_corrected_measure("mean_valence_replies")
        self._calculate_bayesian_corrected_measure("mean_extremity_replies")

    def _calculate_bayesian_corrected_measure(self, measure_name: str):
        """
        Calculate the Bayesian corrected measures for the provided measure name.

        Parameters
        ----------
        measure_name : str
            The name of the measure (in comment data) to bayesian correct.
        """
        weights = self.analytic_functions.get_values_for_quantiles_of_measure(
            self.comment_data[measure_name]
        )

        (
            col_name_q25,
            col_name_q50,
            col_name_q75,
        ) = self._get_bayesian_corrected_column_names(measure_name)

        (
            self.comment_data[col_name_q25],
            self.comment_data[col_name_q50],
            self.comment_data[col_name_q75],
        ) = self.calculate_bayes_corrected_value_for_quantile_weights_vectorized(
            self.comment_data[measure_name],
            self.comment_data["totalvotes"],
            weights,
        )

    @staticmethod
    def _get_bayesian_corrected_column_names(measure_name: str):
        """
        Generate the bayesian corrected column names based on the measure name.

        Parameters:
        measure_name (str): the name of the measure in self.comment_data dictionary

        Returns:
        tuple: The corrected column names for q=0.25, q=0.5, and q=0.75
        """
        if measure_name.startswith("mean_") and measure_name.endswith("_replies"):
            base_name = measure_name.replace("mean_", "").replace("_replies", "")
            prefix = "mean bayes-corrected"
            suffix = " of replies"
        else:
            base_name = measure_name
            prefix = "bayes-corrected"
            suffix = ""

        return (
            f"{prefix} (q=0.25) {base_name}{suffix}",
            f"{prefix} (q=0.5) {base_name}{suffix}",
            f"{prefix} (q=0.75) {base_name}{suffix}",
        )

    def _remove_irrelevant_information_comments(self):
        """
        Remove data from irrelevant columns in the comment data.
        """
        irrelevant_columns = [
            "richTextBody",
            "highlights",
            "id",
            "updated_at",
            "replies",
        ]
        extended_columns = ["body", "status", "parent_id"]

        self.comment_data.drop(columns=irrelevant_columns, inplace=True)
        if not self.add_comments_text_body:
            self.comment_data.drop(columns=extended_columns, inplace=True)

    def _calculate_derived_time_information(
        self, merged_dataset: pd.DataFrame
    ) -> pd.DataFrame:
        """
        Calculate the daytime of comments as well as the time of commenting after
        the release of an article.
        """

        merged_dataset["created_at"] = pd.to_datetime(merged_dataset["created_at"])
        merged_dataset["date_created"] = pd.to_datetime(merged_dataset["date_created"])

        merged_dataset["difference"] = (
            merged_dataset["created_at"] - merged_dataset["date_created"]
        )
        merged_dataset["difference"] = merged_dataset["difference"].dt.total_seconds()
        merged_dataset["comment time after article release (hours)"] = (
            merged_dataset["difference"] / 3600
        )
        merged_dataset["daytime of comment (hours)"] = (
            merged_dataset["created_at"].dt.hour
            + merged_dataset["created_at"].dt.minute / 60
        )

        merged_dataset.drop(
            ["date_created", "created_at", "difference"], axis=1, inplace=True
        )

        return merged_dataset

    def _set_comment_data_datatypes_for_efficient_storing(self):
        """
        Change the datatypes of different columns in comment data to more storage efficient types.
        """
        self.comment_data["order"] = self.comment_data["order"].astype("int16")
        self.comment_data["upvotes"] = self.comment_data["upvotes"].astype("int16")
        self.comment_data["downvotes"] = self.comment_data["downvotes"].astype("int16")
        self.comment_data["number O(n+1)-replies"] = self.comment_data[
            "number O(n+1)-replies"
        ].astype("int16")
        self.comment_data["totalvotes"] = self.comment_data["totalvotes"].astype(
            "int16"
        )
        self.comment_data["valence"] = self.comment_data["valence"].astype("float32")
        self.comment_data["bayes-corrected (q=0.25) valence"] = self.comment_data[
            "bayes-corrected (q=0.25) valence"
        ].astype("float32")
        self.comment_data["bayes-corrected (q=0.5) valence"] = self.comment_data[
            "bayes-corrected (q=0.5) valence"
        ].astype("float32")
        self.comment_data["bayes-corrected (q=0.75) valence"] = self.comment_data[
            "bayes-corrected (q=0.75) valence"
        ].astype("float32")

        self.comment_data["extremity"] = self.comment_data["extremity"].astype(
            "float32"
        )
        self.comment_data["bayes-corrected (q=0.25) extremity"] = self.comment_data[
            "bayes-corrected (q=0.25) extremity"
        ].astype("float32")
        self.comment_data["bayes-corrected (q=0.5) extremity"] = self.comment_data[
            "bayes-corrected (q=0.5) extremity"
        ].astype("float32")
        self.comment_data["bayes-corrected (q=0.75) extremity"] = self.comment_data[
            "bayes-corrected (q=0.75) extremity"
        ].astype("float32")

        self.comment_data["editing.edited"] = (
            self.comment_data["editing.edited"]
            .map({"True": True, "False": False})
            .astype("bool")
        )
        self.comment_data["created_at"] = pd.to_datetime(
            self.comment_data["created_at"]
        )
