"""
The `data_preprocessor.py` module is an abstract base class for data preprocessing.
This class is extended by specific preprocessing classes which implement a `preprocess`
method for specific types of data (e.g., articles or comments).

"""
import os
from abc import ABC, abstractmethod
from multiprocessing import Pool
from tqdm import tqdm
from typing import Optional
import pandas as pd
from pathlib import Path


class DataPreprocessor(ABC):
    """
    An abstract base class used as a template for data preprocessing classes.

    Attributes
    ----------
    data_folder : Path
        Path object representing the directory where the data files are located.
    file_format : str
        The file format extension of the data files (default is ".json").
    split_word : str
        The delimiter used to split filenames to extract article IDs.

    Methods
    -------
    get_all_data() -> Optional[pd.DataFrame]
        Collect all data files from a given directory, preprocess them and return
        a DataFrame containing all processed data.
    preprocess(file_path: Path) -> pd.DataFrame
        Abstract method to preprocess a single file.
        This method should be implemented by specific preprocessing subclasses.
    _get_list_of_article_ids_from_filenames(all_files) -> list[str]
        Get a list of article IDs extracted from the filenames.
    _get_all_files_to_preprocess(data_folder, file_format) -> list[Path]
        Get all files with given format in data_folder recursively.
    """

    def __init__(
        self, data_folder: Path, file_format: str = ".json", split_word: str = ""
    ):
        """
        Constructs the DataPreprocessor object.

        Parameters
        ----------
        data_folder : Path
            Path object representing the directory where the data files are located.
        file_format : str
            The file format extension of the data files (default is ".json").
        split_word : str
            The delimiter used to split filenames to extract article IDs.
        """
        self.data_folder: Path = data_folder
        self.file_format: str = file_format
        self.split_word: str = split_word

    def get_all_data(self) -> Optional[pd.DataFrame]:
        """
        Collect all data files from a given directory, preprocess them and return a DataFrame containing all processed data.

        Returns
        -------
        Optional[pd.DataFrame]
            A DataFrame containing all processed data, or None if no data files were found.
        """
        list_all_files_to_preprocess: list[Path] = self._get_all_files_to_preprocess(
            self.data_folder, self.file_format
        )

        if not list_all_files_to_preprocess:
            raise ValueError(
                "no files found please check the directory path given in the settings"
            )

        list_all_article_ids: list[str] = self._get_list_of_article_ids_from_filenames(
            list_all_files_to_preprocess
        )

        with Pool() as pool:
            list_dataframes: list = []
            for df in tqdm(
                pool.imap_unordered(self.preprocess, list_all_files_to_preprocess),
                total=len(list_all_files_to_preprocess),
            ):
                if df is not None:
                    list_dataframes.append(df)

            if len(list_dataframes) > 0:
                data_all: pd.DataFrame = (
                    pd.concat(list_dataframes, keys=list_all_article_ids)
                    .rename_axis(["article", "idx"])
                    .reset_index()
                )
                data_all: pd.DataFrame = data_all.drop("idx", axis=1)
            else:
                return None

            return data_all

    @abstractmethod
    def preprocess(self, file_path: Path) -> pd.DataFrame:
        """
        Abstract method to preprocess a single file.

        Parameters
        ----------
        file_path : pathlib.Path
            The path object representing the file to be preprocessed.

        Returns
        -------
        pd.DataFrame
            A pandas DataFrame containing the preprocessed data.

        This method should be implemented by specific preprocessing subclasses.
        """
        pass

    def _get_list_of_article_ids_from_filenames(self, all_files) -> list[str]:
        """
        Get a list of article IDs extracted from the filenames.

        Parameters
        ----------
        all_files : List[Path]
            List of Path objects representing the files to be preprocessed.

        Returns
        -------
        List[str]
            List of article IDs.
        """
        list_all_article_ids: list = [
            f.as_posix().replace(self.file_format, "").split(self.split_word, 1)[1]
            for f in all_files
        ]
        return list_all_article_ids

    @staticmethod
    def _get_all_files_to_preprocess(data_folder, file_format) -> list[Path]:
        """
        Get all files with given format in the data_folder path recursively.

        Parameters
        ----------
        data_folder : pathlib.Path
            The path object to the directory where the data files are located.
        file_format : Str
            The file format extension of the data files.

        Returns
        -------
        List[Path]
            List of Path objects representing the files to be preprocessed.
        """
        list_all_files: list = []
        for root, dirs, files in os.walk(os.path.join("..", data_folder)):
            for file in files:
                if file.endswith(file_format):
                    list_all_files.append(Path(root) / file)
        return list_all_files
