"""
The `single_comment_file_preprocessor.py` module provides functionality to preprocess a single comment json file
scraped from Spiegel Online. It includes methods for handling missing data, calculating derived
measures, un-nesting replies, and cleaning the comment data.

"""

import numpy as np
import pandas as pd
from pathlib import Path
from typing import Optional, Tuple

from src.analytic_functions import AnalyticFunctions
from src.utils import load_json_file


class SingleCommentFilePreprocessor:
    """
    A class used to preprocess a single comment file from Spiegel Online.

    Attributes
    ----------
    analytic_functions : AnalyticFunctions
        An instance of AnalyticFunctions used for various analytic computations.
    extract_up_and_downvotes_vectorized : np.vectorize
        Vectorized function to extract upvotes and downvotes.
    calculate_number_of_replies_vectorized : np.vectorize
        Vectorized function to calculate the number of replies.
    calculate_totalvotes_vectorized : np.vectorize
        Vectorized function to calculate the total number of votes.
    calculate_valence_vectorized : np.vectorize
        Vectorized function to calculate the valence of comments.
    calculate_extremity_vectorized : np.vectorize
        Vectorized function to calculate the extremity of comments.

    Methods
    -------
    preprocess(comment_file_path: Path) -> Optional[pd.DataFrame]
        Preprocess the comment data file.
    _get_children_of_comment_entry(comment_entry: dict[str, any]) -> pd.DataFrame
        Get children of a comment entry.
    _extract_up_and_downvotes(action_summary: dict) -> Tuple[int, int]
        Extract upvotes and downvotes from the action summary.
    _handle_missing_users(comments_data: pd.DataFrame) -> pd.DataFrame
        Handle missing user data in the comments.
    _normalize_higher_order(order: int, dataframe_replies: pd.DataFrame, dataframe_ids: pd.DataFrame) -> pd.DataFrame
        Normalize higher order replies.
    _unnest_order(comments_data: pd.DataFrame) -> list[pd.DataFrame]
        Unnest comments by order.
    _unnest_replies_by_order(comments_data: pd.DataFrame) -> pd.DataFrame
        Unnest replies by order.
    _calculate_mean_measure_of_replies(comments_data: pd.DataFrame, column: str) -> pd.DataFrame
        Calculate the mean of a measure (e.g., valence, extremity) for replies.
    _calculate_derived_measures(comments_data: pd.DataFrame) -> pd.DataFrame
        Calculate derived measures like upvotes, downvotes, total votes, valence, extremity, and the number of replies.
    _clean_comments_data(complete_comments_data: pd.DataFrame) -> pd.DataFrame
        Clean the comments data by dropping unnecessary columns.
    """

    def __init__(self):
        """
        Constructs the SingleCommentFilePreprocessor object.
        """
        self.analytic_functions: AnalyticFunctions = AnalyticFunctions()
        self.extract_up_and_downvotes_vectorized = np.vectorize(
            self._extract_up_and_downvotes
        )
        self.calculate_number_of_replies_vectorized = np.vectorize(
            self.analytic_functions.calculate_number_of_replies
        )
        self.calculate_totalvotes_vectorized = np.vectorize(
            self.analytic_functions.calculate_totalvotes
        )
        self.calculate_valence_vectorized = np.vectorize(
            self.analytic_functions.calculate_valence
        )

        self.calculate_extremity_vectorized = np.vectorize(
            self.analytic_functions.calculate_extremity
        )

    def preprocess(self, comment_file_path: Path) -> Optional[pd.DataFrame]:
        """
        Preprocess the comment data file.

        Parameters
        ----------
        comment_file_path : Path
            Path to the JSON file containing the comment data.

        Returns
        -------
        Optional[pd.DataFrame]
            Preprocessed comment data if the file contains comment data, otherwise None.
        """
        comment_data_dict: dict = load_json_file(comment_file_path)

        if not comment_data_dict:
            return None

        comments_data: pd.DataFrame = pd.json_normalize(comment_data_dict)
        comments_data: pd.DataFrame = self._handle_missing_users(comments_data)
        comments_data: pd.DataFrame = self._unnest_replies_by_order(comments_data)
        comments_data: pd.DataFrame = self._calculate_derived_measures(comments_data)
        cleaned_comments_data: pd.DataFrame = self._clean_comments_data(comments_data)

        return cleaned_comments_data

    @staticmethod
    def _get_children_of_comment_entry(comment_entry: dict[str, any]) -> pd.DataFrame:
        """
        Get children of a comment entry.

        Parameters
        ----------
        comment_entry : dict[str, any]
            Dictionary representing a single comment entry.

        Returns
        -------
        pd.DataFrame
            DataFrame containing the children of the comment entry.
        """
        data_frame_children: pd.DataFrame = pd.json_normalize(comment_entry)
        return data_frame_children

    @staticmethod
    def _extract_up_and_downvotes(action_summary: dict) -> Tuple[int, int]:
        """
        Extract upvotes and downvotes from the action summary.

        Parameters
        ----------
        action_summary : dict
            Dictionary containing action summary data.

        Returns
        -------
        Tuple[int, int]
            Upvotes and downvotes count.
        """
        upvotes: int = 0
        downvotes: int = 0
        if len(action_summary) != 0:
            for data in action_summary:
                if data["__typename"] == "UpvoteActionSummary":
                    upvotes: int = data["count"]
                if data["__typename"] == "DownvoteActionSummary":
                    downvotes: int = data["count"]
        return upvotes, downvotes

    @staticmethod
    def _handle_missing_users(comments_data: pd.DataFrame) -> pd.DataFrame:
        """
        Handle missing user data in the comments by creating empty fields to prevent errors.

        Parameters
        ----------
        comments_data : pd.DataFrame
            DataFrame containing the comments data.

        Returns
        -------
        pd.DataFrame
            Modified comments data with missing user data handled.
        """
        if "user.username" not in comments_data:
            comments_data["user.username"]: str = "NaN"
            comments_data["user.id"]: str = "NaN"
            comments_data["user.role"]: str = "NaN"

        return comments_data

    def _normalize_higher_order(
        self, order: int, dataframe_replies: pd.DataFrame, dataframe_ids: pd.DataFrame
    ) -> pd.DataFrame:
        """
        Normalize higher order replies.

        Parameters
        ----------
        order : int
            Current order level of replies.
        dataframe_replies : pd.DataFrame
            DataFrame containing replies.
        dataframe_ids : pd.DataFrame
            DataFrame containing IDs corresponding to higher order replies.

        Returns
        -------
        pd.DataFrame
            Normalized higher order replies.
        """
        list_dataframes_higher_order: list[pd.DataFrame] = [
            self._get_children_of_comment_entry(entry) for entry in dataframe_replies
        ]
        dataframe_higher_order: pd.DataFrame = (
            pd.concat(list_dataframes_higher_order, keys=dataframe_ids)
            .rename_axis(["parent_id", "idx"])
            .reset_index()
        )
        dataframe_higher_order: pd.DataFrame = dataframe_higher_order.drop(
            "idx", axis=1
        )
        dataframe_higher_order["order"]: int = order
        return dataframe_higher_order

    def _unnest_order(self, comments_data: pd.DataFrame) -> list[pd.DataFrame]:
        """
        Unnest comments by order.

        Parameters
        ----------
        comments_data : pd.DataFrame
            DataFrame containing comments data.

        Returns
        -------
        list[pd.DataFrame]
            List of DataFrames with un-nested comments by order.
        """
        list_dataframes_comment_data: list[pd.DataFrame] = [comments_data]
        for i in range(3):
            temporary_dataframe = self._normalize_higher_order(
                i + 1,
                list_dataframes_comment_data[-1]["replies"],
                list_dataframes_comment_data[-1]["id"],
            )
            if temporary_dataframe.get("replies") is not None:
                if temporary_dataframe["replies"].any():
                    list_dataframes_comment_data.append(temporary_dataframe)
            else:
                break
        return list_dataframes_comment_data

    def _unnest_replies_by_order(self, comments_data: pd.DataFrame) -> pd.DataFrame:
        """
        Unnest replies by order.

        Parameters
        ----------
        comments_data : pd.DataFrame
            DataFrame containing comments data.

        Returns
        -------
        pd.DataFrame
            DataFrame with replies un-nested by order.
        """
        comments_data["order"]: int = 0
        comments_data["parent_id"]: int = 0
        list_un_nested_dataframes: list[pd.DataFrame] = self._unnest_order(
            comments_data
        )
        dataframe_full_un_nested_comment_data: pd.DataFrame = (
            pd.concat(list_un_nested_dataframes).reset_index(drop=True)
            if list_un_nested_dataframes
            else comments_data
        )
        return dataframe_full_un_nested_comment_data

    @staticmethod
    def _calculate_mean_measure_of_replies(comments_data: pd.DataFrame, column: str):
        """
        Calculate the mean of a measure (e.g., valence, extremity) for replies.

        Parameters
        ----------
        comments_data : pd.DataFrame
            DataFrame containing the comments data.
        column : str
            The column for which the mean measure of replies is to be calculated.

        Returns
        -------
        pd.DataFrame
            DataFrame with the mean measure of replies added.
        """
        data_replies: pd.DataFrame = (
            comments_data[comments_data["order"] > 0]
            .groupby("parent_id")
            .agg({column: "mean"})
        )

        column_name_mean: str = "mean_" + column + "_replies"

        data_replies.columns = [column_name_mean]
        data_replies.reset_index(inplace=True)
        comments_data: pd.DataFrame = comments_data.merge(
            data_replies, left_on="id", right_on="parent_id", how="left"
        )

        comments_data.drop("parent_id_y", axis=1, inplace=True)
        comments_data.rename(columns={"parent_id_x": "parent_id"}, inplace=True)

        return comments_data

    def _calculate_derived_measures(self, comments_data: pd.DataFrame):
        """
        Calculate measures derived from raw data information like totalvotes, valence, extremity,
        and the number of replies.

        Parameters
        ----------
        comments_data : pd.DataFrame
            DataFrame containing the comments data.

        Returns
        -------
        pd.DataFrame
            DataFrame with derived measures added.
        """
        comments_data["parent_id"]: str = comments_data["parent_id"].astype(str)
        (
            comments_data["upvotes"],
            comments_data["downvotes"],
        ) = self.extract_up_and_downvotes_vectorized(comments_data["action_summaries"])
        comments_data["totalvotes"]: int = self.calculate_totalvotes_vectorized(
            comments_data["upvotes"], comments_data["downvotes"]
        )
        comments_data["valence"]: float = self.calculate_valence_vectorized(
            comments_data["downvotes"], comments_data["totalvotes"]
        )

        comments_data["extremity"]: float = self.calculate_extremity_vectorized(
            comments_data["valence"]
        )

        comments_data["valence"]: float = comments_data["valence"].fillna(0)
        comments_data[
            "number O(n+1)-replies"
        ]: int = self.calculate_number_of_replies_vectorized(comments_data["replies"])

        comments_data: pd.DataFrame = self._calculate_mean_measure_of_replies(
            comments_data, "valence"
        )

        comments_data: pd.DataFrame = self._calculate_mean_measure_of_replies(
            comments_data, "extremity"
        )

        return comments_data

    @staticmethod
    def _clean_comments_data(complete_comments_data):
        """
        Clean the comments data by dropping unnecessary columns.

        Parameters
        ----------
        complete_comments_data : pd.DataFrame
            DataFrame containing the complete comments data.

        Returns
        -------
        pd.DataFrame
            Cleaned DataFrame with unnecessary columns dropped.
        """
        complete_comments_data: pd.DataFrame = complete_comments_data.drop(
            ["action_summaries", "tags", "user.username", "user.role"], axis=1
        )

        if "user" in complete_comments_data:
            complete_comments_data = complete_comments_data.drop("user", axis=1)

        return complete_comments_data
